###############################################################################
# Coscine Python SDK
# Copyright (c) 2018-2021 RWTH Aachen University
# Git: https://git.rwth-aachen.de/coscine/docs/public/coscine-python-sdk
# Please direct bug reports, feature requests or questions at the URL above
# by opening an issue.
#
# Coscine is an open source project at RWTH Aachen University for
# the management of research data.
# Visit https://www.coscine.de for more information.
#
# Please note that this python module resembles a community effort
# and not an official service that RWTH Aachen provides support for.
# It has been initially brough to life by the team at Coscine but has since
# been turned over to the community around Coscine.
###############################################################################

from __future__ import annotations
from typing import TYPE_CHECKING, Callable, List, Type
if TYPE_CHECKING:
	from .client import Client
	from .resource import Resource
import os
import json
from .progressbar import ProgressBar
from .metadata import MetadataForm

###############################################################################

class Object:
	"""
	Objects in Coscine represent file-like data. We could have called it
	File, but in case of linked data we are not actually dealing with files
	themselves, but with links to files. As such, we require a more general
	datatype.
	"""

	client: Client
	resource: Resource
	data: dict
	name: str
	size: int
	type: str
	path: str
	is_folder: bool

	CHUNK_SIZE: int = 4096

###############################################################################

	def __init__(self, resource: Resource, data: dict) -> None:
		"""
		Initializes the Coscine object.

		Parameters
		----------
		resource : coscine.Resource
			Coscine resource handle.
		data : dict
			data of the file-like object.
		"""

		self.client = resource.client
		self.resource = resource
		self.data = data
		self.name = data["Name"]
		self.size = data["Size"]
		self.type = data["Kind"]
		self.path = data["Path"]
		self.is_folder = data["IsFolder"]

###############################################################################

	def __repr__(self) -> str:
		return self.__str__()

###############################################################################

	def __str__(self) -> str:
		return json.dumps(self.data, indent=4)

###############################################################################

	def metadata(self) -> dict:
		"""
		Retrieves the metadata of the file-like object.

		Returns
		-------
		dict
			Metadata as a python dictionary.
		None
			If no metadata has been set for the file-like object.
		"""

		uri = self.client.uri("Tree", "Tree", self.resource.id)
		args = {"path": self.path}
		data = self.client.get(uri, params=args).json()
		metadata = data["data"]["metadataStorage"]
		if not metadata:
			return None
		metadata = metadata[0]
		for key in metadata:
			return metadata[key]

###############################################################################

	def content(self) -> bytes:
		"""
		Retrieves the content/data of the object. In case of linked data
		this would be the link, not the actual file itself. It is impossible
		to get the file contents of linked data objects with this python module.
		In case of rds or rds-s3 data this would return the file contents.
		Be aware that for very large files this will consume a considerable
		amount of RAM memory!

		Returns
		-------
		bytes
			A raw byte-array containing the Coscine file-object's data.
		"""

		uri = self.client.uri("Blob", "Blob", self.resource.id)
		args = {"path": self.path}
		return self.client.get(uri, params = args).content

###############################################################################

	def download(self, path: str = "./", callback: Callable[[int]] = None) -> None:
		"""
		Downloads the file-like object to the local harddrive.

		Parameters
		----------
		path : str, default: "./"
			The path to the download location on the harddrive.
		callback : function(chunksize: int)
			A callback function to be called during downloading chunks.
		"""

		uri = self.client.uri("Blob", "Blob", self.resource.id)
		args = {"path": self.path}
		response = self.client.get(uri, params = args, stream = True)
		path = os.path.join(path, self.name)
		fd = open(path, "wb")
		bar = ProgressBar(self.client, self.size, self.name, "DOWN", callback)
		for chunk in response.iter_content(chunk_size = self.CHUNK_SIZE):
			fd.write(chunk)
			bar.update(len(chunk))
		fd.close()

###############################################################################

	def delete(self) -> None:
		"""
		Deletes the file-like object on the Coscine server.
		"""

		uri = self.client.uri("Blob", "Blob", self.resource.id)
		args = {"path": self.path}
		self.client.delete(uri, params = args)

###############################################################################

	def update(self, metadata) -> None:
		"""
		Updates the metadata of the file-like object.

		Parameters
		----------
		metadata : coscine.MetadataForm or dict
			MetadataForm or JSON-LD formatted metadata dict
		
		Raises
		------
		TypeError
			If argument `metadata` has an unexpected type.
		"""

		if type(metadata) is MetadataForm:
			metadata = metadata.generate()
		elif type(metadata) is not dict:
			raise TypeError("Expected MetadataForm or dict.")
		
		uri = self.client.uri("Tree", "Tree", self.resource.id)
		args = {"path": self.path}
		self.client.put(uri, params = args, data = metadata)

###############################################################################

	def form(self) -> MetadataForm:
		"""
		Returns a MetadataForm filled with the metadata of the file-like object
		"""

		form = MetadataForm(self.resource)
		form.parse(self.metadata())
		return form

###############################################################################

	def objects(self, **kwargs) -> List[Object]:
		if self.is_folder:
			return self.resource.objects(path=self.path, **kwargs)
		else:
			raise TypeError("object is not a directory!")

###############################################################################

	def object(self, displayName: str = None, **kwargs) -> Object:
		if self.is_folder:
			if displayName.endswith("/"): # expect directory
				displayName = self.name + "/" + displayName
			return self.resource.object(displayName = displayName,
											path=self.path, **kwargs)
		else:
			raise TypeError("object is not a directory!")

###############################################################################


###############################################################################
# Coscine Python SDK
# Copyright (c) 2018-2021 RWTH Aachen University
# Git: https://git.rwth-aachen.de/coscine/docs/public/coscine-python-sdk
# Please direct bug reports, feature requests or questions at the URL above
# by opening an issue.
#
# Coscine is an open source project at RWTH Aachen University for
# the management of research data.
# Visit https://www.coscine.de for more information.
#
# Please note that this python module resembles a community effort
# and not an official service that RWTH Aachen provides support for.
# It has been initially brough to life by the team at Coscine but has since
# been turned over to the community around Coscine.
###############################################################################

from typing import Iterable
import urllib
import requests
import json

from .static import StaticServer
from .project import Project, ProjectForm
from .version import __version__
from .banner import BANNER
from .exceptions import *

###############################################################################

class Client:

	"""
	The client class is the backbone and manager of the SDK and mainly
	responsible for the communication and exchange of information 
	with Coscine servers.

	Attributes
	-----------
	lang : str, "en" or "de"
		Language preset for input forms and form data.
	verbose : bool
		Enables/Disables command line output such as log data.
	version : str
		Contains the current version string of the python SDK.
	static : coscine.StaticServer
		Retrieves data from Coscine that rarely/never changes.
	"""

	lang: str
	verbose: bool
	session: requests.Session
	loglevel: Iterable[str]
	version: str
	static: StaticServer

###############################################################################

	def __init__(self, token: str, lang: str = "en", verbose: bool = False,
		loglevel: Iterable[str] = ["LOG", "INFO", "WARN", "REQUEST", "DATA"],\
									persistent_caching: bool = False) -> None:
		"""
		The base class of the Coscine Python SDK. When called it
		creates a new Coscine client object.

		Parameters
		----------
		token : str
			Coscine API access token
		lang : str, "en" or "de", default: "en"
			Language preset for input form fields.
		verbose : bool, default: False
			Enables verbose mode - logging/printing information to stdout.
		loglevel : str
			Enables which messages are printed in verbose mode.
			By default all levels are enabled. Possible values:
			Any combination of values inside an Iterable:
			["LOG", "INFO", "WARN", "REQUEST", "DATA"]
		persistent_caching : bool
			Enable to store the cache in a file on deinitialization of
			the client object. Will attempt to load the cache file
			on initialization if enabled. Leads to a significant speed
			boost when making static requests right after init, but may
			also lead to invalid data when using old cache, in case
			the Coscine API changed. Useful for applications with
			a short runtime which get run often.
		"""

		LANG = ("en", "de")
		if type(token) is not str:
			raise TypeError("Expected argument 'token' as string!")
		if lang not in LANG:
			raise ValueError("Invalid value for argument 'lang'! "\
							"Possible values are %s." % str(LANG))

		self.version = __version__
		self.lang = lang
		self.verbose = verbose
		self.loglevel = loglevel
		self.static = StaticServer(self, persistent_caching)
		self.session = requests.Session()
		self.session.headers = {
			"Authorization": "Bearer " + token,
			"User-Agent": "Coscine Python SDK %s" % self.version
		}
		self.log(BANNER)

###############################################################################

	def log(self, msg: str, level: str = None) -> None:
		"""
		Prints a message to stdout if verbose mode is enabled.

		Parameters
		----------
		msg : str
			Message as a string.
		level :	str, default: None
			Which loglevel to use for printing.
			Possible values: "LOG", "INFO", "WARN",	"REQUEST", "DATA", None
		"""

		LOGLEVELS = ["LOG", "INFO", "WARN", "REQUEST", "DATA"]

		if not self.verbose:
			return
		if level is None:
			print(msg)
		elif self.verbose and level in LOGLEVELS:
			print("[%s] %s" % (level, msg))
		else:
			raise ValueError("Invalid value for argument 'level'! "\
						"Possible values are %s." % str(LOGLEVELS))

###############################################################################

	def latest_version(self) -> str:
		"""
		Retrieves the version string of the latest version of this
		package hosted on PyPi. Useful for checking whether the currently
		used version is outdated and if an update should be performed.

		Examples
		--------
		>>> if client.version != client.latest_version():
		>>> 	print("Module outdated.")
		>>> 	print("Run 'py -m pip install --upgrade coscine'.")
		```

		Returns
		-------
		str
			The version string of the Coscine Python SDK hosted on PyPi.
		"""

		URI = "https://pypi.org/pypi/coscine/json"
		data = requests.get(URI).json()
		version = data["info"]["version"]
		return version

###############################################################################

	@staticmethod
	def uri(api: str, endpoint: str, *args) -> str:
		"""
		Constructs a URL for performing a request to the Coscine API.

		Parameters
		----------
		api : str
			The target Coscine API endpoint.
		endpoint : str
			The subendpoint of `api`.
		*args
			Variable number of arguments of type string to append to the URL.
			Arguments are automatically seperated by a slash '/' and
			special characters are encoded.
		
		Returns
		-------
		str
			Encoded URL for communicating with Coscine servers.
		"""

		BASE = "https://coscine.rwth-aachen.de/coscine/api/Coscine.Api.%s/%s"
		ENDPOINTS = (
			"Blob",
			"Metadata",
			"Organization",
			"Project",
			"Resources",
			"Tree",
			"User",
			"ActivatedFeatures"
		)

		if api not in ENDPOINTS:
			raise ValueError("Invalid value for argument 'api'. "\
						"Possible values are %s." % str(ENDPOINTS))

		uri = BASE % (api, endpoint)
		for arg in args:
			if arg is None:
				continue
			uri += "/" + urllib.parse.quote(arg, safe="")

		return uri

###############################################################################

	def _request(self, method: str, uri: str, **kwargs) -> requests.Response:
		"""
		Performs a HTTP request to the Coscine Servers.

		Parameters
		----------
		method : str
			HTTP request method.
		uri : str
			Coscine URL generated with Client.uri(...).
		**kwargs
			Additional keyword arguments forwarded to the requests library.
		
		Raises
		------
		coscine.exceptions.ConnectionError
			If the Coscine servers could not be reached.
		coscine.exceptions.AuthorizationError
			If the Coscine API token is invalid.
		coscine.exceptions.ClientError
			If the request resulted in an error.

		Returns
		-------
		requests.Response
			The response of the Coscine server as a requests.Response object.
		"""

		self.log("%s %s" % (method, uri), "REQUEST")

		if "data" in kwargs and type(kwargs["data"]) is dict:
			kwargs["headers"] = {
				"Content-Type": "application/json;charset=utf-8"
			}
			kwargs["data"] = json.dumps(kwargs["data"])
			self.log(json.dumps(kwargs["data"], indent=4), "DATA")

		try:
			response = self.session.request(method, uri, **kwargs)
			response.raise_for_status()
			return response
		except requests.exceptions.ConnectionError:
			raise ConnectionError()
		except requests.exceptions.RequestException as e:
			if e.response.status_code == 401:
				raise AuthorizationError("Invalid API token!")
			else:
				raise ClientError()

###############################################################################

	def get(self, uri: str, **kwargs) -> requests.Response:
		"""
		Performs a GET request to the Coscine API.

		Parameters
		----------
		uri : str
			Coscine URL generated with Client.uri(...).
		**kwargs
			Additional keyword arguments forwarded to the requests library.
		
		Examples
		--------
		>>> uri = client.uri("Project", "Project")
		>>> projects = client.get(uri).json()

		Raises
		------
		coscine.exceptions.ConnectionError
			If the Coscine servers could not be reached.
		coscine.exceptions.AuthorizationError
			If the Coscine API token is invalid.
		coscine.exceptions.ClientError
			If the request resulted in an error.

		Returns
		-------
		requests.Response
			The response of the Coscine server as a requests.Response object.
		"""

		return self._request("GET", uri, **kwargs)

###############################################################################

	def put(self, uri: str, **kwargs) -> requests.Response:
		"""
		Performs a PUT request to the Coscine API.

		Parameters
		----------
		uri : str
			Coscine URL generated with Client.uri(...).
		**kwargs
			Additional keyword arguments forwarded to the requests library.
		
		Examples
		--------
		>>> uri = self.client.uri("Tree", "Tree", resource.id, filename)
		>>> self.client.put(uri, data = metadata)

		Raises
		------
		coscine.exceptions.ConnectionError
			If the Coscine servers could not be reached.
		coscine.exceptions.AuthorizationError
			If the Coscine API token is invalid.
		coscine.exceptions.ClientError
			If the request resulted in an error.

		Returns
		-------
		requests.Response
			The response of the Coscine server as a requests.Response object.
		"""

		return self._request("PUT", uri, **kwargs)

###############################################################################

	def post(self, uri: str, **kwargs) -> requests.Response:
		"""
		Performs a POST request to the Coscine API.

		Parameters
		----------
		uri : str
			Coscine URL generated with Client.uri(...).
		**kwargs
			Additional arguments forwarded to the requests library.
		
		Examples
		--------
		TODO

		Raises
		------
		coscine.exceptions.ConnectionError
			If the Coscine servers could not be reached.
		coscine.exceptions.AuthorizationError
			If the Coscine API token is invalid.
		coscine.exceptions.ClientError
			If the request resulted in an error.

		Returns
		-------
		requests.Response
			The response of the Coscine server as a requests.Response object.
		"""

		return self._request("POST", uri, **kwargs)

###############################################################################

	def delete(self, uri: str, **kwargs) -> requests.Response:
		"""
		Performs a DELETE request to the Coscine API.

		Parameters
		----------
		uri : str
			Coscine URL generated with Client.uri(...).
		**kwargs
			Additional keyword arguments forwarded to the requests library.
		
		Examples
		--------
		TODO

		Raises
		------
		coscine.exceptions.ConnectionError
			If the Coscine servers could not be reached.
		coscine.exceptions.AuthorizationError
			If the Coscine API token is invalid.
		coscine.exceptions.ClientError
			If the request resulted in an error.

		Returns
		-------
		requests.Response
			The response of the Coscine server as a requests.Response object.
		"""

		return self._request("DELETE", uri, **kwargs)

###############################################################################

	def projects(self, toplevel: bool = True, **kwargs) -> Iterable[Project]:
		"""
		Retrieves a list of a all projects the creator of the Coscine API token
		is currently a member of.

		Parameters
		----------
		toplevel : bool, default: True
			Retrieve only toplevel projects (no subprojects).
			Set to False if you want to retrieve all projects, regardless
			of hierarchy.
		**kwargs
			Project filter values.
			-> e.g. displayName="MyProject"
			-> Returns all projects with the specified displayName

		Returns
		-------
		list
			List of coscine.Project objects
		"""

		ENDPOINTS = ("Project", "Project/-/topLevel")
		uri = self.uri("Project", ENDPOINTS[toplevel])
		projects = []
		for entry in self.get(uri).json():
			for key, value in kwargs.items():
				if entry[key] != value:
					break
			else:
				projects.append(Project(self, entry))
		return projects

###############################################################################

	def project(self, displayName: str = None, toplevel: bool = True, \
												**kwargs) -> Project:
		"""
		Retrieves a single project the holder of the Coscine
		API token is currently a member of. Search criteria
		must be specified.

		Parameters
		----------
		toplevel : bool, default: True
			Search only within toplevel projects (no subprojects)
			Set to False if you want to search through all projects,
			regardless of hierarchy.
		**kwargs (mandatory!)
			Project filter values.
			-> e.g. displayName="MyProject"
			-> Returns one project matching the specified displayName

		Raises
		-------
		coscine.exceptions.AmbiguityError
			If more than one project match the specified search criteria.

		Returns
		-------
		coscine.Project
			coscine.Project object if a project matches the specifed filter.
		None
			if no project matching the filter was found.
		"""

		if displayName:
			kwargs["displayName"] = displayName
		elif not kwargs:
			raise ParameterError("Not enough arguments!")
		projects = self.projects(toplevel, **kwargs)
		if len(projects) == 1:
			return projects[0]
		elif len(projects) == 0:
			return None
		else:
			raise AmbiguityError("Found more than 1 project matching "\
											"the specified criteria!")

###############################################################################

	def create_project(self, form: ProjectForm) -> Project:
		"""
		Creates a project using the given ProjectForm.

		Parameters
		----------
		form : coscine.ProjectForm
			ProjectForm filled with project metadata.

		Returns
		-------
		coscine.Project
			Project object of the new project.
		"""

		if type(form) is ProjectForm:
			form = form.generate()
		
		uri = self.uri("Project", "Project")
		return Project(self, self.post(uri, data=form).json())

###############################################################################

	def ProjectForm(self, parent: Project = None) -> ProjectForm:
		"""
		Generates a project form to create and edit project metadata.

		Parameters
		----------
		parent : coscine.Project, default: None
			A parent Coscine project.
		
		Returns
		--------
		coscine.ProjectForm
		"""

		return ProjectForm(self, parent)

###############################################################################


###############################################################################
# Coscine Python SDK
# Copyright (c) 2018-2021 RWTH Aachen University
# Git: https://git.rwth-aachen.de/coscine/docs/public/coscine-python-sdk
# Please direct bug reports, feature requests or questions at the URL above
# by opening an issue.
#
# Coscine is an open source project at RWTH Aachen University for
# the management of research data.
# Visit https://www.coscine.de for more information.
#
# Please note that this python module resembles a community effort
# and not an official service that RWTH Aachen provides support for.
# It has been initially brough to life by the team at Coscine but has since
# been turned over to the community around Coscine.
###############################################################################

from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
	from .client import Client

import os
import json
import urllib
import atexit
from .version import __version__

###############################################################################

class StaticServer:

	"""
	The purpose of the StaticServer class is to provide
	static data which is not subject to change. Initially a request is made
	to the Coscine REST API to query such data. Upon further requests a
	cached response is used to speed things up and reduce internet traffic.
	Dynamic data such as the metadata of a project cannot be cached as
	it may be invalidated by further requests or interactions with a
	different client such as the web interface.
	"""

	client: Client
	cache: dict
	persist: bool

	CACHEFILE = "./.coscine_python_sdk_%s-cache.json" % __version__

	def __init__(self, client: Client, persist: bool = False) -> None:
		"""
		Intializes the StaticServer with a Coscine SDK client handle.
		"""

		self.client = client
		self.persist = persist
		if persist:
			atexit.register(self.savestate)
			try:
				fd = open(self.CACHEFILE, "r")
				self.cache = json.loads(fd.read())
				fd.close()
			except FileNotFoundError:
				self.cache = {}
		else:
			self.cache = {}

###############################################################################

	def savestate(self) -> None:
		"""
		Saves the cache if the persist option is enabled
		"""

		if self.persist and self.cache:
			fd = open(self.CACHEFILE, "w")
			fd.write(json.dumps(self.cache))
			fd.close()

###############################################################################

	def disciplines(self) -> dict:
		"""
		Queries the scientific disciplines available for selection in Coscine.

		Returns
		-------
		dict
			Dictionary containing the disciplines as keys.
		"""

		LANG = {
			"de": "displayNameDe",
			"en": "displayNameEn"
		}

		uri = self.client.uri("Project", "Discipline")
		if uri in self.cache:
			data = self.cache[uri]
		else:
			data = self.client.get(uri).json()
			self.cache[uri] = data
		
		disciplines = {}
		lang = LANG[self.client.lang]
		for entry in data:
			disciplines[entry[lang]] = entry
		
		return disciplines

###############################################################################

	def organizations(self) -> dict:
		"""
		Queries the organizations (e.g. 'RWTH Aachen University') available
		for selection in Coscine.

		Returns
		-------
		dict
			A python dictionary containing the organizations as keys.
		"""

		uri = self.client.uri("Organization", "Organization")
		if uri in self.cache:
			data = self.cache[uri]
		else:
			data = self.client.get(uri).json()
			self.cache["organizations"] = data

		organizations = {}
		for entry in data["data"]:
			organizations[entry["displayName"]] = entry
		return organizations

###############################################################################

	def visibility(self) -> dict:
		"""
		Gets the key-value mapping of the Coscine project visibility field.

		Returns
		-------
		dict
			Key-value mapped Coscine project visibility field.
		"""

		uri = self.client.uri("Project", "Visibility")
		if uri in self.cache:
			data = self.cache[uri]
		else:
			data = self.client.get(uri).json()
			self.cache[uri] = data

		visibility = {}
		for entry in data:
			visibility[entry["displayName"]] = entry
		return visibility

###############################################################################

	def features(self) -> dict:
		"""
		Returns a mapping for activated features of Coscine projects.
		Activated features resemble for example the announcement board
		and the documents section.

		Returns
		-------
		dict
			Dictionary containing the available features.
		"""

		uri = self.client.uri("ActivatedFeatures", "ActivatedFeatures")
		if uri in self.cache:
			data = self.cache[uri]
		else:
			data = self.client.get(uri).json()
			self.cache[uri] = data

		features = {}
		for entry in data:
			features[entry[self.client.lang]] = entry
		return features

###############################################################################

	def licenses(self) -> dict:
		"""
		Returns a dictionary containing a list of licenses available in Coscine
		as keys. The dictionary is supposed to be used for mapping license names
		to their internal representation in Coscine.

		Returns
		-------
		dict
		"""

		uri = self.client.uri("Project", "License")
		if uri in self.cache:
			data = self.cache[uri]
		else:
			data = self.client.get(uri).json()
			self.cache[uri] = data

		licenses = {}
		for entry in data:
			licenses[entry["displayName"]] = entry
		return licenses

###############################################################################

	def resource_types(self) -> dict:
		"""
		Retrieves a list of the available resource types in Coscine.

		Returns
		-------
		dict
		"""

		uri = self.client.uri("Resources", "ResourceType", "types")
		if uri in self.cache:
			data = self.cache[uri]
		else:
			data = self.client.get(uri).json()
			self.cache[uri] = data

		types = {}
		for it in data:
			if it["isEnabled"]:
				types[it["displayName"]] = it
		return types

###############################################################################

	def application_profiles(self) -> dict:
		"""
		Queries the list of available application profiles in Coscine.

		Returns
		--------
		dict
			A dictionary containing the application profile names as keys.
		"""

		uri = self.client.uri("Metadata", "Metadata", "profiles")
		if uri in self.cache:
			data = self.cache[uri]
		else:
			data = self.client.get(uri).json()
			self.cache[uri] = data

		profiles = {}
		for entry in data:
			name = urllib.parse.urlparse(entry)[2]
			name = os.path.relpath(name, "/coscine/ap/")
			name = name.upper()
			profiles[name] = entry
		return profiles

###############################################################################

	@staticmethod
	def _parse_application_profile(profile: dict) -> dict:
		"""
		Parses an application profile into an easier to access dictionary by
		altering the key naming. It should increase readability both of the
		application profile itself and the code accessing/modifying it.

		Parameters
		-----------
		profile : dict
			The application profile to parse as a dictionary. Retrieved by
			e.g. StaticServer.application_profile(...).

		Returns
		-------
		dict
			The modified/parsed application profile as a dictionary
		"""

		def _get_lang(entry, lang):
			for it in entry:
				if it["@language"] == lang:
					return it
			return None

		# This function is absolutely cursed, isn't it? :^)
		# Don't stare at it for too long, or it may haunt you in your dreams.

		W3PREFIX = "http://www.w3.org/ns/shacl#%s"
		data = {}
		profile = profile[0]
		data["id"] = profile["@id"]
		graph = []
		for entry in profile["@graph"]:
			obj = {}
			if W3PREFIX % "name" not in entry:
				continue
			obj["id"] = entry["@id"]
			obj["path"] = entry[W3PREFIX % "path"][0]["@id"]
			obj["order"] = int(entry[W3PREFIX % "order"][0]["@value"])
			if W3PREFIX % "minCount" in entry:
				obj["minCount"] = int(entry[W3PREFIX % "minCount"][0]["@value"])
			if W3PREFIX % "maxCount" in entry:
				obj["maxCount"] = int(entry[W3PREFIX % "maxCount"][0]["@value"])
			obj["name"] = {
				"de": _get_lang(entry[W3PREFIX % "name"], "de")["@value"],
				"en": _get_lang(entry[W3PREFIX % "name"], "en")["@value"]
			}
			if W3PREFIX % "datatype" in entry:
				obj["datatype"] = entry[W3PREFIX % "datatype"][0]["@id"]
				obj["type"] = "literal"
			if W3PREFIX % "class" in entry:
				obj["class"] = entry[W3PREFIX % "class"][0]["@id"]
				obj["datatype"] = obj["class"]
				obj["type"] = "uri"
			graph.append(obj)
		data["graph"] = graph
		return data

###############################################################################

	def application_profile(self, path: str, id: str = None,
								parse: bool = False) -> dict:
		"""
		Retrieves a specific application profile.

		Parameters
		-----------
		path : str
			Path/Url to the application profile.
			(e.g. Resource.data["applicationProfile"]))
		id : str, default: None
			Coscine resource ID.
		parse : bool
			Specifies wether the application profile should be parsed with
			self._parse_application_profile(...) before returning.
		
		Returns
		-------
		dict
			An application profile as a dictionary.
		"""

		uri = self.client.uri("Metadata", "Metadata", "profiles", path, id)
		if uri in self.cache:
			profile = self.cache[uri]
		else:
			profile = self.client.get(uri).json()
			self.cache[uri] = profile

		if parse:
			profile = self._parse_application_profile(profile)
		return profile

###############################################################################

	def instance(self, id: str, link: str) -> dict:
		"""
		Returns a static project instance

		Parameters
		----------
		id : str
			Coscine project id.
		link : str
			uri to query.
		

		Returns
		-------
		dict
			instance
		"""

		uri = self.client.uri("Metadata", "Metadata", "instances", id, link)
		if uri in self.cache:
			data = self.cache[uri]
		else:
			data = self.client.get(uri).json()
			self.cache[uri] = data
		
		return data

###############################################################################


###############################################################################
# Coscine Python SDK
# Copyright (c) 2018-2021 RWTH Aachen University
# Git: https://git.rwth-aachen.de/coscine/docs/public/coscine-python-sdk
# Please direct bug reports, feature requests or questions at the URL above
# by opening an issue.
#
# Coscine is an open source project at RWTH Aachen University for
# the management of research data.
# Visit https://www.coscine.de for more information.
#
# Please note that this python module resembles a community effort
# and not an official service that RWTH Aachen provides support for.
# It has been initially brough to life by the team at Coscine but has since
# been turned over to the community around Coscine.
###############################################################################

from __future__ import annotations
from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
	from .client import Client
	from . import ProjectForm

import os
import json
from prettytable import PrettyTable

from .metadata import MetadataPresetForm
from .exceptions import *
from .resource import Resource, ResourceForm
from .form import InputForm

###############################################################################

class Project:
	"""
	Python representation of a Coscine Project
	"""

	client: Client
	data: dict
	parent: Project
	id: str
	name: str
	displayName: str
	description: str
	principleInvestigators: str
	startDate: str
	endDate: str
	disciplines: List[str]
	organizations: List[str]
	visibility: str

###############################################################################

	def __init__(self, client: Client, data: dict, parent: Project = None) -> None:
		"""
		Initializes a project object.

		Parameters
		----------
		client : coscine.Client
			Coscine client handle
		data : dict
			Project data received from Coscine.
		parent : coscine.Project
			Optional parent project.
		"""

		self.client = client
		self.data = data
		self.parent = parent
		self.id = data["id"]
		self.name = data["projectName"]
		self.displayName = data["displayName"]
		self.description = data["description"]
		self.principleInvestigators = data["principleInvestigators"]
		self.startDate = data["startDate"]
		self.endDate = data["endDate"]
		self.disciplines = [k["displayNameEn"] for k in data["disciplines"]]
		self.organizations = [k["displayName"] for k in data["organizations"]]
		self.visibility = data["visibility"]["displayName"]

###############################################################################

	def __repr__(self) -> str:
		return self.__str__()

###############################################################################

	def __str__(self) -> str:
		table = PrettyTable(("Field", "Value"))
		rows = [
			("id", self.id),
			("Name", self.name),
			("Display Name", self.displayName),
			("Principle Investigators", self.principleInvestigators),
			("Disciplines", "\n".join(self.disciplines)),
			("Organizations", "\n".join(self.organizations)),
			("Start date", self.startDate),
			("End date", self.endDate)
		]
		table.max_width["Value"] = 50
		table.add_rows(rows)
		return table.get_string(title="Project")

###############################################################################

	def subprojects(self, **kwargs) -> List[Project]:
		"""
		Retrieves a list of subprojects of the current project matching a set
		of specified filters

		Parameters
		----------
		kwargs
			key-value filters for subprojects (e.g. 'Name' = 'My Project').

		Returns
		-------
		list of coscine.Project
		"""

		uri = self.client.uri("Project", "SubProject", self.id)
		projects = self.client.get(uri).json()
		filter = []
		for data in projects:
			match = True
			for key, value in kwargs.items():
				if data[key] != value:
					match = False
					break
			if match:
				filter.append(Project(self.client, data, self))
		return filter

###############################################################################

	def resources(self, **kwargs) -> List[Resource]:
		"""
		Retrieves a list of Resources of the current project matching a set
		of specified filters.

		Parameters
		----------
		kwargs
			key-value filters for resources (e.g. 'Name' = 'My Resource').
		
		Returns
		-------
		list[coscine.Resource]
			list of resources matching the supplied filter.
		"""

		uri = self.client.uri("Project", "Project", self.id, "resources")
		resources = []
		for data in self.client.get(uri).json():
			for key, value in kwargs.items():
				if data[key] != value:
					break
			else:
				resources.append(Resource(self, data))
		return resources

###############################################################################

	def resource(self, displayName: str = None, **kwargs) -> Resource:
		"""
		Retrieves a certain resource of the current project matching a set
		of specified filters or identified by its displayName.

		Parameters
		----------
		displayName : str
			The display name of the resource.
		kwargs
			key-value filters for resources (e.g. 'Name' = 'My Resource').
		
		Returns
		--------
		coscine.Resource or None
		"""
		
		if displayName:
			kwargs["displayName"] = displayName
		elif not kwargs:
			raise ParameterError("")

		resources = self.resources(**kwargs)
		if len(resources) == 1:
			return resources[0]
		elif len(resources) == 0:
			return None
		else:
			raise AmbiguityError("Found more than 1 resource matching "\
											"the specified criteria!")

###############################################################################

	def download(self, path: str = "./", metadata: bool = False) -> None:
		"""
		Downloads the project to the location referenced by 'path'.

		Parameters
		----------
		path : str
			Download location on the harddrive
			Default: current directory './'
		metadata : bool, default: False
			If enabled, project metadata is downloaded and put in
			a hidden file '.metadata.json'.
		"""

		path = os.path.join(path, self.displayName)
		if not os.path.isdir(path):
			os.mkdir(path)
		for resource in self.resources():
			resource.download(path=path, metadata=metadata)
		if metadata:
			data = json.dumps(self.data, indent=4)
			fd = open(os.path.join(path, ".metadata.json"), "w")
			fd.write(data)
			fd.close()

###############################################################################

	def delete(self) -> None:
		"""
		Deletes the project on the remote server.
		"""

		uri = self.client.uri("Project", "Project", self.id)
		self.client.delete(uri)

###############################################################################

	def members(self) -> List[ProjectMember]:
		"""
		Retrieves a list of all members of the current project

		Returns
		--------
		list of coscine.project.ProjectMember
			List of project members as ProjectMember objects.
		"""

		uri = self.client.uri("Project", "ProjectRole", self.id)
		data = self.client.get(uri).json()
		members = [ProjectMember(self, m) for m in data]
		return members

###############################################################################

	def invite(self, email: str, role: str = "Member") -> None:
		"""
		Invites a person to a project via their email address

		Parameters
		----------
		email : str
			The email address to send the invite to
		role : str, "Member" or "Owner", default: "Member"
			The role for the new project member
		"""

		ROLES = {
			"Owner": "be294c5e-4e42-49b3-bec4-4b15f49df9a5",
			"Member": "508b6d4e-c6ac-4aa5-8a8d-caa31dd39527"
		}

		if role not in ROLES:
			raise ValueError("Invalid role '%s'." % role)

		uri = self.client.uri("Project", "Project", "invitation")
		data = {
			"projectId": self.data["id"],
			"role": ROLES[role],
			"email": email
		}

		try:
			self.client.log("Inviting [%s] as [%s] to project [%s]." % 
												(email, role, self.id))
			self.client.post(uri, data = data)
		except ServerError:
			self.client.log("User [%s] has invite pending." % email)

###############################################################################

	def create_resource(self, form: ResourceForm, metadataPreset: \
							MetadataPresetForm = None) -> Resource:
		"""
		Creates a resource within the current project using the supplied
		resource form.

		Parameters
		----------
		resourceForm : coscine.ResourceForm
			Form to generate the resource with.
		metadataPreset : coscine.MetadataPresetForm
			optional application profile configuration.
		"""
		if type(form) is ResourceForm:
			form = form.generate()
		if metadataPreset:
			if type(metadataPreset) is MetadataPresetForm:
				metadataPreset = metadataPreset.generate()
			form["fixedValues"] = metadataPreset
		uri = self.client.uri("Resources", "Resource", "Project", self.id)
		return Resource(self, self.client.post(uri, data = form).json())

###############################################################################

	def form(self) -> ProjectForm:
		"""
		Returns a ProjectForm filled with metadata of the current project.
		"""
		form = ProjectForm(self.client)
		form.parse(self.data)
		return form

###############################################################################

	def update(self, form: ProjectForm) -> None:
		"""
		Updates a project using the given ProjectForm

		Parameters
		----------
		form : coscine.ProjectForm
			ProjectForm containing updated data.
		"""

		if type(form) is ProjectForm:
			form = form.generate()
		uri = self.client.uri("Project", "Project", self.id)
		self.client.post(uri, data = form)

###############################################################################

	def ResourceForm(self, data: dict = None) -> ResourceForm:
		"""
		Creates a ResourceForm for this project.

		Returns
		-------
		coscine.ResourceForm
		"""

		return ResourceForm(self, data)

###############################################################################

class ProjectMember:

	"""
	Initializes a project member for a given project.

	Parameters
	----------
	project : coscine.Project
		Coscine python SDK project handle.
	data: dict
		User data as dict, retrieved via 
		client.uri("Project", "ProjectRole", self.id).
	"""

	name: str
	email: str
	id: str
	role: str
	data: dict
	client: Client
	project: Project

	def __init__(self, project: Project, data: dict) -> None:
		"""
		"""

		self.project = project
		self.client = self.project.client
		self.data = data
		self.name = data["user"]["displayName"]
		self.email = data["user"]["emailAddress"]
		self.id = data["user"]["id"]
		self.role = data["role"]["displayName"]

###############################################################################

	def set_role(self, role: str) -> None:

		"""
		Sets the role of a project member

		Parameters
		----------
		role : str
			The new role of the member ('Owner' or 'Member').
		"""

		ROLES = {
			"Owner": "be294c5e-4e42-49b3-bec4-4b15f49df9a5",
			"Member": "508b6d4e-c6ac-4aa5-8a8d-caa31dd39527"
		}

		if role not in ROLES:
			raise ValueError("Invalid role '%s'." % role)

		uri = self.client.uri("Project", "ProjectRole")
		self.data["role"]["id"] = ROLES[role]
		self.data["role"]["displayName"] = role
		self.client.post(uri, data = self.data)

###############################################################################

	def remove(self) -> None:
		"""
		Removes a project member from their associated project.
		"""

		uri = self.client.uri("Project", "ProjectRole", "project", \
			self.project.id, "user", self.id, "role", self.data["role"]["id"])
		self.client.delete(uri)

###############################################################################

KEYS = [{
		"name": {
			"de": "Projektname",
			"en": "Project Name"
		},
		"flags": InputForm.REQUIRED,
		"field": "ProjectName"
	},{
		"name": {
			"de": "Anzeigename",
			"en": "Display Name"
		},
		"flags": InputForm.REQUIRED,
		"field": "DisplayName"
	},{
		"name": {
			"de": "Projektbeschreibung",
			"en": "Project Description"
		},
		"flags": InputForm.REQUIRED,
		"field": "Description"
	},{
		"name": {
			"de": "Principal Investigators",
			"en": "Principal Investigators"
		},
		"flags": InputForm.REQUIRED,
		"field": "PrincipleInvestigators"
	},{
		"name": {
			"de": "Projektstart",
			"en": "Project Start"
		},
		"flags": InputForm.REQUIRED,
		"field": "StartDate"
	},
	{
		"name": {
			"de": "Projektende",
			"en": "Project End"
		},
		"flags": InputForm.REQUIRED,
		"field": "EndDate"
	},{
		"name": {
			"de": "Disziplin",
			"en": "Discipline"
		},
		"flags": InputForm.REQUIRED | InputForm.CONTROLLED | InputForm.LIST,
		"field": "Discipline"
	},{
		"name": {
			"de": "Teilnehmende Organisation",
			"en": "Participating Organizations"
		},
		"flags": InputForm.REQUIRED | InputForm.CONTROLLED | InputForm.LIST,
		"field": "Organization"
	},{
		"name": {
			"de": "Projektschlagwörter",
			"en": "Project Keywords"
		},
		"flags": InputForm.NONE,
		"field": "Keywords"
	},{
		"name": {
			"de": "Grant ID",
			"en": "Grant ID"
		},
		"flags": InputForm.NONE,
		"field": "GrantId"
	},{
		"name": {
			"de": "Features",
			"en": "Features"
		},
		"flags": InputForm.CONTROLLED,
		"field": "Features"
}]

###############################################################################

MAP = {
	"description": "Description",
	"displayName": "DisplayName",
	"startDate": "StartDate",
	"endDate": "EndDate",
	"keywords": "Keywords",
	"projectName": "ProjectName",
	"principleInvestigators": "PrincipleInvestigators",
	"grantId": "GrantId",
	"disciplines": "Discipline",
	"organizations": "Organization"
}

###############################################################################

class ProjectForm(InputForm):

	"""
	The Coscine input form for interacting with projects. Set and modify
	project	settings.
	"""

	def __init__(self, client: Client, parent: Project = None, data: dict = None) -> None:
		self.client = client
		vocabulary = {
			"Discipline": self.client.static.disciplines(),
			"Organization": self.client.static.organizations(),
			"Visibility": self.client.static.visibility(),
			"Features": self.client.static.features()
		}
		super().__init__("Project Form", client.lang, KEYS, vocabulary)
		self.parent = parent
		# Fill
		if data:
			for key in data:
				self[key] = data[key]

###############################################################################

	def parse(self, data: dict):

		"""
		Parses given project data into a project form.

		Parameters
		-----------
		data : dict
			Dict containing Coscine project metadata (i.e. Project.data).
		"""

		def _parse_disciplines(self, data):
			LSTR = {"en": "displayNameEn", "de": "displayNameDe"}[self._lang]
			disciplines = []
			for discipline in data:
				disciplines.append(discipline[LSTR])
			return disciplines

		def _parse_organizations(self, data):
			organizations = []
			for organization in data:
				vocabulary = self._vocabulary["Organization"]
				for entry in vocabulary.values():
					if entry["url"] == organization["url"]:
						organizations.append(entry["displayName"])
			return organizations

		for key in data:
			if key in MAP and data[key] != "":
				entry = self.entry(MAP[key])
				if entry is None:
					continue
				name = entry["name"][self._lang]
				flags = entry["flags"]
				value = data[key]
				if flags & InputForm.CONTROLLED or flags & InputForm.LIST:
					if key == "disciplines":
						self[name] = _parse_disciplines(self, value)
					elif key == "organizations":
						self[name] = _parse_organizations(self, value)
					elif key == "visibility":
						self[name] = value["displayName"]
				else:
					self[name] = value

###############################################################################

	def generate(self) -> dict:

		"""
		Generates Coscine JSON-LD representation of form data. Such data
		can be sent to Coscine for processing.

		Returns
		-------
		dict
			JSON-LD metadata dict
		"""

		def final_value(key, value):
			if self.is_controlled(key):
				return self._vocabulary[field][value]
			else:
				return value

		data = {}
		missing = []
		for key in self._keys:
			entry = self._keys[key]
			field = entry["field"]
			if key not in self.store:
				if self.is_required(key):
					missing.append(key)
			else:
				value = self.store[key]
				if entry["flags"] & ProjectForm.LIST:
					if type(value) is not list and type(value) is not tuple:
						raise ValueError("Expected iterable value for key %s" % key)
					data[field] = []
					for v in value:
						data[field].append(final_value(key, v))
				else:
					data[field] = final_value(key, value)

		if missing:
			raise RequirementError(missing)
		
		if self.parent:
			data["ParentId"] = self.parent.id

		data["Visibility"] = {
			"displayName": "Project Members",
			"id": "8ab9c883-eb0d-4402-aaad-2e4007badce6"
		}

		return data

###############################################################################


###############################################################################
# Coscine Python SDK
# Copyright (c) 2018-2021 RWTH Aachen University
# Git: https://git.rwth-aachen.de/coscine/docs/public/coscine-python-sdk
# Please direct bug reports, feature requests or questions at the URL above
# by opening an issue.
#
# Coscine is an open source project at RWTH Aachen University for
# the management of research data.
# Visit https://www.coscine.de for more information.
#
# Please note that this python module resembles a community effort
# and not an official service that RWTH Aachen provides support for.
# It has been initially brough to life by the team at Coscine but has since
# been turned over to the community around Coscine.
###############################################################################

###############################################################################
# This file contains a printable banner. It is printed to stdout
# upon initialization of the SDK with verbose mode enabled.
# The banner includes the current package version and title. Otherwise
# it serves no special purpose and can be disregarded for anything other
# than demonstrations.
###############################################################################

from .version import __version__

###############################################################################

BANNER = \
"""
                     _             
                    (_)            
   ___ ___  ___  ___ _ _ __   ___  
  / __/ _ \/ __|/ __| | '_ \ / _ \ 
 | (_| (_) \__ \ (__| | | | |  __/ 
  \___\___/|___/\___|_|_| |_|\___| 
____________________________________

    Coscine Python SDK %s
____________________________________
""" % __version__

###############################################################################


###############################################################################
# Coscine Python SDK
# Copyright (c) 2018-2021 RWTH Aachen University
# Git: https://git.rwth-aachen.de/coscine/docs/public/coscine-python-sdk
# Please direct bug reports, feature requests or questions at the URL above
# by opening an issue.
#
# Coscine is an open source project at RWTH Aachen University for
# the management of research data.
# Visit https://www.coscine.de for more information.
#
# Please note that this python module resembles a community effort
# and not an official service that RWTH Aachen provides support for.
# It has been initially brough to life by the team at Coscine but has since
# been turned over to the community around Coscine.
###############################################################################

# Do not set version to 1.0.0 before Coscine end of pilot phase
__version__ = "0.5.3"


###############################################################################
# Coscine Python SDK
# Copyright (c) 2018-2021 RWTH Aachen University
# Git: https://git.rwth-aachen.de/coscine/docs/public/coscine-python-sdk
# Please direct bug reports, feature requests or questions at the URL above
# by opening an issue.
#
# Coscine is an open source project at RWTH Aachen University for
# the management of research data.
# Visit https://www.coscine.de for more information.
#
# Please note that this python module resembles a community effort
# and not an official service that RWTH Aachen provides support for.
# It has been initially brough to life by the team at Coscine but has since
# been turned over to the community around Coscine.
###############################################################################

from __future__ import annotations
from typing import TYPE_CHECKING, Callable, List
if TYPE_CHECKING:
	from .client import Client
	from .project import Project
import os
import json
from prettytable import PrettyTable
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor
from .object import Object
from .form import InputForm
from .metadata import MetadataForm
from .exceptions import *
from .progressbar import ProgressBar

###############################################################################

class Resource:
	"""
	Coscine Resource type.

	Attributes
	----------
	id : str
		Coscine-internal resource-ID
	pid : str
		Persistent Identifier of the resource
	"""

	client: Client
	project: Project
	data: dict
	id: str
	pid: str
	name: str
	displayName: str
	type: str
	disciplines: List[str]
	profile: str
	archived: bool
	creator: str
	s3: S3info

###############################################################################

	class S3info:
		access_key: str = ""
		secret_key: str = ""
		endpoint: str	= ""
		bucket: str		= ""
		
		def __init__(self, data: dict) -> None:
			if "resourceTypeOption" in data and data["resourceTypeOption"]:
				self.access_key = data["resourceTypeOption"].get("WriteAccessKey")
				self.secret_key = data["resourceTypeOption"].get("WriteSecretKey")
				self.endpoint = data["resourceTypeOption"].get("Endpoint")
				self.bucket = data["resourceTypeOption"].get("BucketName")

###############################################################################

	def __init__(self, project: Project, data: dict) -> None:
		"""
		Initializes the Resource object

		Parameters
		----------
		project : coscine.Project
			Coscine project handle
		data : dict
			Coscine resource data.
		"""

		self.client = project.client
		self.project = project
		self.data = data
		self.id = data["id"]
		self.pid = data["pid"]
		self.name = data["resourceName"]
		self.displayName = data["displayName"]
		self.license = data["license"]["displayName"] \
						if self.data["license"] else ""
		self.type = self.data["type"]["displayName"]
		self.disciplines = [k["displayNameEn"] for k in data["disciplines"]]
		self.profile = self.data["applicationProfile"]
		self.archived = self.data["archived"]
		self.creator = self.data["creator"]
		self.s3 = Resource.S3info(data)

###############################################################################

	def __repr__(self) -> str:
		return self.__str__()

###############################################################################

	def __str__(self) -> str:
		table = PrettyTable(["Field", "Value"])
		rows = [
			("ID", self.id),
			("Resource Name", self.name),
			("Display Name", self.displayName),
			("PID", self.pid),
			("Type", self.type),
			("Disciplines", "\n".join(self.disciplines)),
			("License", self.license),
			("Application Profile", self.profile),
			("Archived", self.archived),
			("Creator", self.creator)
		]
		table.max_width["Value"] = 50
		table.add_rows(rows)
		return table.get_string(title = "Resource")

###############################################################################

	def upload(self, key: str, file, metadata: dict = None, \
					callback: Callable[[int]] = None) -> None:
		"""
		Uploads a file-like object to a resource on the Coscine server

		Parameters
		----------
		key : str
			filename of the file-like object.
		file : object with read() attribute
				Either open file handle or local file location path.
		metadata : dict
			File metadata. For rds-s3 this is optional, but recommended.
		callback : Callable[int]
			Optional callback called during chunk uploads.
		"""

		if hasattr(file, "read"):
			fd = file
			filename = "MEM"
		elif type(file) is str:
			fd = open(file, "rb")
			filename = file
		else:
			raise TypeError("Argument `file` has unexpected type!")

		if metadata:
			if type(metadata) is MetadataForm:
				metadata = metadata.generate()
			uri = self.client.uri("Tree", "Tree", self.id, key)
			self.client.put(uri, data = metadata)

		uri = self.client.uri("Blob", "Blob", self.id, key)
		fields = {"files": (key, fd, "application/octect-stream")}
		encoder = MultipartEncoder(fields = fields)
		bar = ProgressBar(self.client, encoder.len, filename, "UP", callback)
		monitor = MultipartEncoderMonitor(encoder, callback = lambda monitor: \
										bar.update(monitor.bytes_read - bar.n))
		headers = {"Content-Type": monitor.content_type}
		self.client.put(uri, data = monitor, headers = headers)

###############################################################################

	def download(self, path: str = "./", metadata: bool = False) -> None:
		"""
		Downloads the resource and all of its contents to the local harddrive.

		Parameters
		----------
		path : str, default: "./"
			Path to the local storage location.
		metadata : bool, default: False
			If enabled, resource metadata is downloaded and put in
			a hidden file '.metadata.json'.
		"""

		path = os.path.join(path, self.name)
		if not os.path.isdir(path):
			os.mkdir(path)
		for file in self.objects():
			file.download(path)
		if metadata:
			data = json.dumps(self.data, indent=4)
			fd = open(os.path.join(path, ".resource-metadata.json"), "w")
			fd.write(data)
			fd.close()

###############################################################################

	def delete(self) -> None:
		"""
		Deletes the Coscine resource and all objects contained within it.
		"""

		uri = self.client.uri("Resources", "Resource", self.id)
		self.client.delete(uri)

###############################################################################

	def quota(self) -> int:
		"""
		Retrieves the used up quota of the resource.

		Returns
		-------
		int
			Used up quota in bytes.
		"""

		uri = self.client.uri("Blob", "Blob", self.id, "quota")
		data = self.client.get(uri).json()
		return int(data["data"]["usedSizeByte"])

###############################################################################

	def objects(self, path: str = None, **kwargs) -> List[Object]:
		"""
		Returns a list of Objects stored within the resource

		Parameters
		------------
		kwargs
			file-object filter arguments (e.g. 'Name' = 'testfile').

		Returns
		-------
		list[Object]
			List of Coscine file-like objects.
		"""

		objects = []
		uri = self.client.uri("Tree", "Tree", self.id)
		if path:
			args = {"path": path}
		else:
			args = None
		data = self.client.get(uri, params = args).json()
		fileStorage = data["data"]["fileStorage"]
		metadataStorage = data["data"]["metadataStorage"]
		for data in fileStorage:
			for key, value in kwargs.items():
				if data[key] != value:
					break
			else:
				objects.append(Object(self, data))
		return objects

###############################################################################

	def object(self, displayName: str = None, path: str = None, **kwargs) \
																 -> Object:
		"""
		Returns an Object stored within the resource

		Parameters
		------------
		displayName : str, default: None
			file-object display name (filename/key).
		kwargs
			Filter

		Returns
		-------
		coscine.Object
			Python representation of the file-object as an Object instance
		"""
		
		if displayName:
			if displayName.endswith("/"):
				displayName = displayName[:-1]
			kwargs["Name"] = displayName
		elif not kwargs:
			raise ParameterError("")

		objects = self.objects(path=path, **kwargs)
		if len(objects) == 1:
			return objects[0]
		elif len(objects) == 0:
			return None
		else:
			raise AmbiguityError("Found more than 1 resource matching "\
											"the specified criteria!")

###############################################################################

	def application_profile(self, parse: bool = False) -> dict:
		"""
		Returns the application profile of the resource

		Parameters
		----------
		parse : bool, default: False
			Parse the profile into a more readable format.

		Returns
		--------
		dict
			Application profile of the resource, either parsed or raw.
		"""

		return self.client.static.application_profile(\
						self.profile, self.id, parse)

###############################################################################

	def form(self) -> ResourceForm:
		"""
		Returns a ResourceForm filled with the metadata of the current resource.

		Returns
		-------
		coscine.resource.ResourceForm
		"""

		form = ResourceForm(self.project)
		form.parse(self.data)
		return form

###############################################################################

	def update(self, form: ResourceForm) -> None:
		"""
		Updates the metadata of the resource using the supplied ResourceForm.

		Parameters
		----------
		form : coscine.resource.ResourceForm
			ResourceForm filled with updated values.
		"""

		if type(form) is ResourceForm:
			form = form.generate()
		elif type(form) is not dict:
			raise TypeError("")

		uri = self.client.uri("Resources", "Resource", self.id)
		self.client.post(uri, data = form)

###############################################################################

	def set_archived(self, flag: bool) -> None:
		"""
		Set the archived flag of the resource to put it in read-only mode.
		Only the resource creator or project owner can do this.

		Parameters
		----------
		flag : bool
			Enable with True, Disable with False.
		"""

		uri = self.client.uri("Resources", "Resource", self.id, \
					"setReadonly?status=%s" % str(flag).lower())
		self.client.post(uri)
		#self.data["archived"] = flag
		self.archived = flag

###############################################################################

	def MetadataForm(self, data: dict = None) -> MetadataForm:
		"""
		Creates a MetadataForm for this resource.

		Returns
		-------
		coscine.MetadataForm
		"""

		return MetadataForm(self, data)

###############################################################################

KEYS = [{
		"name": {
			"de": "Ressourcentyp",
			"en": "Resource Type"
		},
		"flags": InputForm.REQUIRED | InputForm.CONTROLLED,
		"field": "type"
	},{ # This is REQUIRED for rds and s3, but not for Linked Data!
		"name": {
			"de": "Ressourcengröße",
			"en": "Resource Size"
		},
		"flags": InputForm.SPECIAL,
		"field": "resourceTypeOption"
	},{
		"name": {
			"de": "Ressourcenname",
			"en": "Resource Name"
		},
		"flags": InputForm.REQUIRED,
		"field": "ResourceName"
	},{
		"name": {
			"de": "Anzeigename",
			"en": "Display Name"
		},
		"flags": InputForm.REQUIRED,
		"field": "DisplayName"
	},{
		"name": {
			"de": "Ressourcenbeschreibung",
			"en": "Resource Description"
		},
		"flags": InputForm.REQUIRED,
		"field": "Description"
	},{
		"name": {
			"de": "Disziplin",
			"en": "Discipline"
		},
		"flags": InputForm.REQUIRED | InputForm.CONTROLLED | InputForm.LIST,
		"field": "Disciplines"
	},{
		"name": {
			"de": "Ressourcenschlagwörter",
			"en": "Resource Keywords"
		},
		"flags": InputForm.NONE,
		"field": "Keywords"
	},{
		"name": {
			"de": "Lizenz",
			"en": "License"
		},
		"flags": InputForm.CONTROLLED,
		"field": "License"
	},{
		"name": {
			"de": "Verwendungsrechte",
			"en": "Usage Rights"
		},
		"flags": InputForm.NONE,
		"field": "UsageRights"
	},{
		"name": {
			"de": "Applikationsprofile",
			"en": "Application Profile"
		},
		"flags": InputForm.REQUIRED | InputForm.CONTROLLED,
		"field": "applicationProfile"
}]

MAP = {
	"description": "Description",
	"displayName": "DisplayName",
	"resourceName": "ResourceName",
	"keywords": "Keywords",
	"disciplines": "Disciplines",
	"license": "License",
	"resourceTypeOption": "resourceTypeOption",
	"applicationProfile": "applicationProfile",
	"type": "type",
	"usageRights": "UsageRights"
}

###############################################################################

class ResourceForm(InputForm):

	"""
	Coscine Input Form for creating and editing resources
	"""

###############################################################################

	def __init__(self, project: Project, data: dict = None) -> None:

		"""
		Initializes the Resource form. The fields are static and fixed.

		Parameters
		-----------
		project : coscine.Project
			Coscine project handle.
		"""

		vocabulary = {
			"type": project.client.static.resource_types(),
			"applicationProfile": project.client.static.application_profiles(),
			"License": project.client.static.licenses(),
			"Visibility": project.client.static.visibility(),
			"Disciplines": project.client.static.disciplines()
		}

		super().__init__("Resource Form", project.client.lang, KEYS, vocabulary)
		# Fill
		if data:
			for key in data:
				self[key] = data[key]

###############################################################################

	def parse(self, data: dict) -> None:

		"""
		Parses given resource data into a resource form.

		Parameters
		-----------
		data : dict
			Dict containing Coscine resource metadata (i.e. Resource.data).
		"""

		def _parse_disciplines(self, data):
			LSTR = {"en": "displayNameEn", "de": "displayNameDe"}[self._lang]
			disciplines = []
			for discipline in data:
				disciplines.append(discipline[LSTR])
			return disciplines

		for key in data:
			if key in MAP and data[key] != "":
				entry = self.entry(MAP[key])
				if entry is None:
					continue
				name = entry["name"][self._lang]
				flags = entry["flags"]
				value = data[key]
				if flags & InputForm.CONTROLLED or \
				flags & InputForm.LIST or \
				flags & InputForm.SPECIAL:
					if key == "disciplines":
						self[name] = _parse_disciplines(self, value)
					elif key == "resourceTypeOption":
						self[name] = value["Size"]
					elif key == "applicationProfile":
						values = list(self._vocabulary["applicationProfile"].values())
						keys = list(self._vocabulary["applicationProfile"].keys())
						position = values.index(value)
						self[name] = keys[position]
					elif value is not None and (key == "visibility"\
					 or key == "license" or key == "type"):
						self[name] = value["displayName"]
				else:
					self[name] = value

###############################################################################

	def generate(self) -> dict:

		"""
		Generates JSON-LD formatted representation of resource data

		Raises
		-------
		RequirementError
			When one or more required fields have not been set
		
		Returns
		--------
		dict
			JSON-LD formatted resource data
		"""

		def final_value(key, value):
			if self.is_controlled(key):
				return self._vocabulary[field][value]
			else:
				return value

		data = {}
		missing = []
		for key in self._keys:
			entry = self._keys[key]
			field = entry["field"]
			if key not in self.store:
				if self.is_required(key):
					missing.append(key)
			else:
				value = self.store[key]
				if entry["flags"] & ResourceForm.LIST:
					if type(value) is not list and type(value) is not tuple:
						raise ValueError("Expected iterable value for key %s" % key)
					data[field] = []
					for v in value:
						data[field].append(final_value(key, v))
				else:
					data[field] = final_value(key, value)

		if missing:
			raise RequirementError(missing)

		size = {}
		if "resourceTypeOption" in data:
			size = {
				"Size": data["resourceTypeOption"]
			}
		elif data["type"]["displayName"] in ("rds", "rds-s3"):
			raise RequirementError("rds or rds-s3 require a size parameter!")
		data["resourceTypeOption"] = size

		data["Visibility"] = {
			"displayName": "Project Members",
			"id": "8ab9c883-eb0d-4402-aaad-2e4007badce6"
		}

		return data

###############################################################################


###############################################################################
# Coscine Python SDK
# Copyright (c) 2018-2021 RWTH Aachen University
# Git: https://git.rwth-aachen.de/coscine/docs/public/coscine-python-sdk
# Please direct bug reports, feature requests or questions at the URL above
# by opening an issue.
#
# Coscine is an open source project at RWTH Aachen University for
# the management of research data.
# Visit https://www.coscine.de for more information.
#
# Please note that this python module resembles a community effort
# and not an official service that RWTH Aachen provides support for.
# It has been initially brough to life by the team at Coscine but has since
# been turned over to the community around Coscine.
###############################################################################

from __future__ import annotations
from typing import TYPE_CHECKING, Callable
if TYPE_CHECKING:
	from .client import Client
import tqdm

###############################################################################

class ProgressBar:
	"""
	The ProgressBar class is a simple wrapper around tqdm
	progress bars. It is used in download/upload methods and provides the
	benefit of remembering state information and printing only when in
	verbose mode.

	Attributes
	----------
	client : coscine.Client
		Coscine Python SDK client handle.
	bar : tqdm.tqdm
		tqdm Progress Bar instance.
	n : int
		Number of bytes read.
	callback : Callable[int]
		Callback function to call on update.
	"""

	client: Client
	bar: tqdm.tqdm
	n: int
	callback: Callable[[int]]

	def __init__(self, client: Client, filesize: int, key: str, \
				mode: str, callback: Callable[[int]] = None) -> None:
		"""
		Initializes a state-aware tqdm ProgressBar.

		client : coscine.Client
			Coscine python SDK client handle
		filesize : int
			Size of the file object in bytes
		key : str
			key/filename of the file object
		mode : str
			'UP' or 'DOWN' referring to upload and download
		callafter : function(chunksize: int)
			 callback function to call after each update
		"""

		MODES = {
			"UP":	"↑",
			"DOWN":	"↓"
		}

		if mode not in MODES:
			raise ValueError("Invalid value for argument 'mode'! "\
					"Possible values are %s." % str(MODES.keys()))
		
		self.client = client
		self.callback = callback
		self.n = 0
		if self.client.verbose:
			self.bar = tqdm.tqdm(total=filesize, unit="B", unit_scale=True,\
										desc="%s %s" % (MODES[mode], key))

###############################################################################

	def update(self, chunksize: int) -> None:
		"""
		Updates the progress bar with respect to the consumed chunksize.
		If a callafter function has been provided to the Constructor, it is
		called during the update.

		Parameters
		----------
		chunksize : int
			indicates the amount of bytes consumed since the last update
		"""

		self.n += chunksize
		if self.client.verbose:
			self.bar.update(chunksize)
		if self.callback:
			self.callback(chunksize)

###############################################################################


###############################################################################
# Coscine Python SDK
# Copyright (c) 2018-2021 RWTH Aachen University
# Git: https://git.rwth-aachen.de/coscine/docs/public/coscine-python-sdk
# Please direct bug reports, feature requests or questions at the URL above
# by opening an issue.
#
# Coscine is an open source project at RWTH Aachen University for
# the management of research data.
# Visit https://www.coscine.de for more information.
#
# Please note that this python module resembles a community effort
# and not an official service that RWTH Aachen provides support for.
# It has been initially brough to life by the team at Coscine but has since
# been turned over to the community around Coscine.
###############################################################################

###############################################################################
# This file defines exceptions used by the Coscine python SDK.
# The base exception class is called CoscineException. It directly inherits
# its properties from the standard python Exception class.
# Although never used directly it provides the basis for more specific
# exceptions, thus allowing you to catch any Coscine exception with one
# except statement.
###############################################################################

class CoscineException(Exception):
	"""
	Coscine base Exception class.
	"""
	pass

###############################################################################

class ConnectionError(CoscineException):
	"""
	In case the client is not able to establish a connection with
	the Coscine servers, a ConnectionError is raised.
	"""
	pass

###############################################################################

class ClientError(CoscineException):
	"""
	An error has been made or detected on the client side.
	"""
	pass

###############################################################################

class ServerError(CoscineException):
	"""
	An error has been made or detected on the Coscine server.
	"""
	pass

###############################################################################

class VocabularyError(CoscineException):
	"""
	Raised in InputForms when a supplied value is not contained within
	a controlled vocabulary.
	"""
	pass

###############################################################################

class RequirementError(CoscineException):
	"""
	Commonly raised in InputForms when a required field has not been set.
	"""
	pass

###############################################################################

class AuthorizationError(CoscineException):
	"""
	AuthorizationErrors are thrown when the owner of the Coscine API
	token does not hold enough privileges.
	"""
	pass

###############################################################################

class AmbiguityError(CoscineException):
	"""
	An AmbiguityError is raised in cases where two objects could
	not be differentiated between.
	"""
	pass

###############################################################################

class ParameterError(CoscineException):
	"""
	Invalid (number of) function parameters provided. In some cases
	the user has the option of choosing between several optional arguments,
	but has to provide at least one.
	"""
	pass

###############################################################################